using System;
using System.IO;
using Zcu.Mve.Core;
using System.Windows.Forms;
using System.ComponentModel;
using Zcu.Mve.Strings;

namespace Zcu.Gsvd.MeshLS
{
	/// <summary>
	/// Main class of MeshLoader. Loads and saves configuration file and runs file readers.
	/// </summary>
	[ModuleInfo("Jan Brta", "Module for loading of mesh from .tri, .stl, .ply and .cmx file formats.",
		        IconName="MeshLS.ico", Assemblied = "2005-5-24", Category = "File Loader")]
	[PortInfo("Output", "Zcu.Mve.Visualization.Mesh - contains triangle mesh (tri, stl, cmx format) or polygons (ply format).")]
	public class MeshLoader : Module
	{
		/// <summary>
		/// Stores type of input file.
		/// </summary>
		internal byte typeOfFile = 0;

		/// <summary>
		/// Stores name of input file.
		/// </summary>
		internal string nameOfFile = null;

		[Editor(typeof(System.Windows.Forms.Design.FileNameEditor), typeof(System.Drawing.Design.UITypeEditor))]
		public string FileName
		{
			get { return nameOfFile; }
			set
			{
				byte type;

				if (!File.Exists(value))          // Test if the file exists.
				{
					MessageBox.Show("Invalid file name!\n\nClick Cancel or choose existing file.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);					
				}

				if ((type = this.tellFileType(value.ToLower())) == 0)
				{                                              // Test the extension of file - 0 is code for unsupported type.
					MessageBox.Show("Unknown file format!\n\nSee help for list of supported file types.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);					
				}

				this.nameOfFile = value;
				// Change meshLoader file name property.
				this.typeOfFile = type;             // Change meshLoader file type property.

				this.newFileSelected = true;    // ... set flag that cause generating new triangle mesh.

			}
		}

		/// <summary>
		/// Indicates that new file (filename) is selected.
		/// </summary>
		internal bool newFileSelected = true;

		/// <summary>
		/// Reference of DataSet class that contains mesh data.
		/// </summary>
		private DataSet dataSet = null;

		/// <summary>
		/// Mesh data that is on output port of module.
		/// </summary>
		private Zcu.Mve.Visualization.Mesh mesh = null;

		/// <summary>
		/// Constructor of MeshLoader. Creates output port of module.
		/// </summary>
		public MeshLoader()
		{
			AddOutPort("Output", typeof (Zcu.Mve.Visualization.Mesh));
			AddInPort("FileName", typeof(Zcu.Mve.Strings.MveString), false);
		}

		/// <summary>
		/// Entry point of module. Reads input file and sets data on output port.
		/// </summary>
		public override void Execute()
		{
			if (this.GetInPort("FileName").Link != null)
			{
				if (!IsAllDataSame())
				{
					this.FileName = ((MveString)GetInput("FileName")).Value;
				}
			}
			if ((this.newFileSelected) && this.typeOfFile != 0)
			{
				this.newFileSelected = false;

				try 
				{
					if (!System.IO.File.Exists(this.nameOfFile))
					{
						throw new FileNotFoundException("");
					}

					int startTics = Environment.TickCount;
					this.ReadFile();
					int endTics   = Environment.TickCount;

					Console.WriteLine("File loaded in {0} ms", (endTics - startTics));
				}
				catch(MveException re)
				{
					this.newFileSelected = true;

					throw re;
				}

				catch(FileNotFoundException)
				{
					this.newFileSelected = true;

					throw new MveException("Input file not found.");
				}

				catch(System.Exception)
				{
					this.newFileSelected = true;
					throw new MveException("Fatal error in meshLoader - uncatch System exception.");
				}

				this.mesh = this.dataSet.GetMesh();        // All is OK - loaded mesh is on output port.
				SetOutput("Output", mesh);
			}

		}


		// LVA 06/07/06 - removed in order to use the srandard setup dialog
		/*
		/// <summary>
		/// Creates MeshLoaderSetup, where user can sets input file.
		/// </summary>
		/// <returns>Module setup.</returns>
		public override ModuleSetup InvokeSetup()
		{
			return new MeshLoaderSetup(this, this.nameOfFile);
		}*/

		/// <summary>
		/// Configuration is saved to XML map file.
		/// </summary>
		/// <param name="config">Element of XML that will be saved in map file.</param>
		public override void WriteConfig(System.Xml.XmlElement config)
		{
			System.Xml.XmlElement xmlEl;

			xmlEl = config.OwnerDocument.CreateElement("Output_file");
			xmlEl.SetAttribute("path", this.nameOfFile);
			xmlEl.SetAttribute("type", this.typeOfFile.ToString());
			config.AppendChild(xmlEl);
		}

		/// <summary>
		/// Read of configuration from XML map file.
		/// </summary>
		/// <param name="config">Element of XML that is saved in map file.</param>
		public override void ReadConfig(System.Xml.XmlElement config)
		{
			System.Xml.XmlNodeList childList = config.ChildNodes;

			try                                            // Try to read specified attributes.
			{
				if ((this.nameOfFile = childList[0].Attributes["path"].Value) == "")
				{
					this.nameOfFile = null;
				}

				this.typeOfFile = Byte.Parse(childList[0].Attributes["type"].Value);
			}
			catch
			{                                              // No config data saved in map file.
				this.nameOfFile = null;
				this.typeOfFile = 0;
			}
		}

		/// <summary>
		/// Analyze type of selected file based on file extension.
		/// </summary>
		/// <param name="path">Full filename (full path and filename).</param>
		/// <returns>Return type of selected input file - 1 - TRI format,
		///                                               2 - STL format,
		///                                               3 - PLY format,
		///                                               4 - CMX format,
		///												  5 - WRL format,
		///                                               0 - ??? format.</returns>
		private byte tellFileType(string path)
		{
			if (path.EndsWith(".tri"))
			{
				return 1;
			}
			else if (path.EndsWith(".stl"))
			{
				return 2;
			}
			else if (path.EndsWith(".ply"))
			{
				return 3;
			}
			else if (path.EndsWith(".cmx"))
			{
				return 4;
			}
			else if (path.EndsWith(".wrl"))
			{
				return 5;
			}
			else
			{
				return 0;
			}
		}

		/// <summary>
		/// Call methods for read input file based on type of file.
		/// </summary>
		private void ReadFile() 
		{
			switch (this.typeOfFile) 
			{
				case 1: this.ReadTriFile();
					break;
				case 2: this.ReadStlFile();
					break;
				case 3: this.ReadPlyFile();
					break;
				case 4: this.ReadCmxFile();
					break;
				case 5: this.ReadWrlFile();
					break;

			}
		}

		/// <summary>
		/// Creates reader object of TRI format.
		/// </summary>
		private void ReadTriFile()
		{
			ReaderTri reader = new ReaderTri();            // Creates file reader.

			this.dataSet = reader.ReadFile(this.nameOfFile);
			                                               // Reads file.
		}

		/// <summary>
		/// Creates reader object of WRL format.
		/// </summary>
		private void ReadWrlFile()
		{
			ReaderWrl reader = new ReaderWrl();            // Creates file reader.

			this.dataSet = reader.ReadFile(this.nameOfFile);
			// Reads file.
		}

		/// <summary>
		/// Analyze input file with STL extension and decide if it's binary or text.
		/// </summary>
		private void ReadStlFile()
		{
			StreamReader streamReader = new StreamReader(this.nameOfFile);

			string line = streamReader.ReadLine().ToLower();

			if (line.StartsWith("solid"))
			{
				this.ReadAsciiStlFile();
			}
			else
			{
				this.ReadBinStlFile();
			}
		}

		/// <summary>
		/// Creates reader object of text STL format.
		/// </summary>
		private void ReadAsciiStlFile()
		{
			ReaderStl reader = new ReaderStl();            // Creates file reader.

			this.dataSet = reader.ReadFile(this.nameOfFile);
			                                               // Reads file.
		}

		/// <summary>
		/// Creates reader object of binary STL format.
		/// </summary>
		private void ReadBinStlFile()
		{
			ReaderBinStl reader = new ReaderBinStl();      // Creates file reader.

			this.dataSet = reader.ReadFile(this.nameOfFile);
			                                               // Reads file.
		}

		/// <summary>
		/// Creates reader object of PLY format.
		/// </summary>
		private void ReadPlyFile()
		{
			ReaderPly reader = new ReaderPly();            // Creates file reader.

			this.dataSet = reader.ReadFile(this.nameOfFile);
                                                           // Reads file.
		}

		/// <summary>
		/// Creates reader object of CMX format.
		/// </summary>
		private void ReadCmxFile()
		{
			ReaderCmx reader = new ReaderCmx();            // Creates file reader.

			this.dataSet = reader.ReadFile(this.nameOfFile);
                                                           // Reads file.
		}
	}
}
