using System;
using System.IO;
using System.Text;

namespace Zcu.Gsvd.MeshLS
{
	/// <summary>
	/// Class for reading PLY file format.
	/// </summary>
	public class ReaderPly
	{
		private int numberOfCells = 0;
		private int numberOfVertices  = 0;

		private int actualLine = 0;                        // Actual line of readed file.

		private DataSet dataSet;

		private string exception = null;                   // Exception error string.

		/// <summary>
		/// Reads input file.
		/// </summary>
		/// <param name="fileName">Full filename.</param>
		/// <returns>Filled object of DataSet class.</returns>
		public DataSet ReadFile(string fileName)
		{
			StreamReader streamReader = new StreamReader(fileName);

			string line = null;
			byte   numberOf = 2;                           //  Counting parts of file head.

			try
			{
				while ((line = streamReader.ReadLine()) != null) 
				{                                          // File head processing.
					this.actualLine++;

					if (line.StartsWith("element ver"))
					{                                      // Reads number of vertices.
						this.numberOfVertices = this.IntNumber(line);
						numberOf--;
					}
					else if (line.StartsWith("element fac"))
					{                                      // Reads number of cells.
						this.numberOfCells = this.IntNumber(line);
						numberOf--;
					}

					if (numberOf == 0) 
					{                                      // Head reading is done, creating DataSet object.
						this.dataSet = new DataSet(this.numberOfVertices, this.numberOfCells, 1);
						break;
					}
				}

				if (numberOf != 0)                             // If number of triangles or vertices wasn't found - throws exception.
				{
					this.exception = "Invalid file format - can't find number of triangles and/or number of vertices.";
					throw new Zcu.Mve.Core.MveException(null);
				}

				while ((line = streamReader.ReadLine()) != null)
				{                                              // Finds end of file head.
					this.actualLine++;

					if (line.StartsWith("end_header"))
					{
						break;
					}
				}

				this.ReadVertices(ref streamReader);
				this.ReadPolygons(ref streamReader);
			}
			catch (System.Exception)
			{
				streamReader.Close();
				if (this.exception == null)
				{
					throw new Zcu.Mve.Core.MveException("Error in input file on line " + this.actualLine + ".");
				}
				else
				{
					throw new Zcu.Mve.Core.MveException(this.exception);
				}
			}

			streamReader.Close();

			return this.dataSet;
		}

		/// <summary>
		/// Reads integer number in input string.
		/// </summary>
		/// <param name="line">Input string.</param>
		/// <returns>Readed number.</returns>
		private int IntNumber(string line) 
		{
			string[] elements;

			elements = line.Split(" :\t".ToCharArray());

			return Int32.Parse(elements[elements.Length - 1]);
		}

		/// <summary>
		/// Reads one vertex in input string.
		/// </summary>
		/// <param name="line">Input string.</param>
		/// <returns>Readed vertex (3D point).</returns>
		private Zcu.Mve.Visualization.Point3D ReadPoint3D(string line) 
		{
			string[] elements;
			int elementsLength = 0;

			System.Globalization.NumberFormatInfo nfi = new System.Globalization.NumberFormatInfo();

			nfi.CurrencyDecimalSeparator = ".";            // Setting of decimal separator.

			elements = line.Split(" :\t".ToCharArray());

			for (int i = 0; i < elements.Length; i++)      // Delete elements that contains spaces only.
			{
				if (elements[i].Trim() != "")
				{
					elements[elementsLength++] = elements[i];
				}
			}

			return new Zcu.Mve.Visualization.Point3D(Double.Parse(elements[0], nfi),
				                                     Double.Parse(elements[1], nfi),
				                                     Double.Parse(elements[2], nfi));
		}

		/// <summary>
		/// Reads one polygon in input string.
		/// </summary>
		/// <param name="line">Input string.</param>
		/// <returns>Readed polygon.</returns>
		private Zcu.Mve.Visualization.Polygon ReadPolygon(string line) 
		{
			string[] elements;
			int[] polygon;
			int elementsLength = 0;
			int polyDimension;

			elements = line.Split(" :\t".ToCharArray());

			for (int i = 0; i < elements.Length; i++)      // Deletes elements that contains spaces only.
				if (elements[i].Trim() != "")
					elements[elementsLength++] = elements[i];


			polyDimension = Int32.Parse(elements[0]);      // Reads number of polygon vertices.
			polygon = new int[polyDimension];              // Allocate memory for polygon.

			for (int i = 0; i < polyDimension; i++)        // Reads indexes of polygon vertices.
			{
				polygon[i] = Int32.Parse(elements[i + 1]);
			}

			return new Zcu.Mve.Visualization.Polygon(polygon);
		}

		/// <summary>
		/// Reads all vertices from input file.
		/// </summary>
		/// <param name="streamReader">Reference to input file.</param>
		private void ReadVertices(ref StreamReader streamReader)
		{
			Zcu.Mve.Visualization.Point3D vertex;
			string   line;
			int    i = 0;                                  // Index of actual vertex.

			while ((line = streamReader.ReadLine()) != null) 
			{
				this.actualLine++;

				line = line.Trim();

				if (line == "")                            // Skips free line.
				{
					continue;
				}
				else                                       // Trys to read one vertex.
				{
					vertex = this.ReadPoint3D(line);
					this.dataSet.AddVertex(i, vertex);

					if (++i == numberOfVertices)           // Number of vertices reached - exit.
					{
						break;
					}
				}
			}			
		}

		/// <summary>
		/// Reads all polygons from input file.
		/// </summary>
		/// <param name="streamReader">Reference to input file.</param>
		private void ReadPolygons(ref StreamReader streamReader)
		{
			Zcu.Mve.Visualization.Polygon polygon;
			string   line;
			int    i = 0;                                  // Index of actual triangle.

			while ((line = streamReader.ReadLine()) != null)
			{
				this.actualLine++;

				line = line.Trim();

				if (line == "")                            // Skips free line.
				{
					continue;
				}
				else                                       // Try to reads one triangle.
				{
					polygon = this.ReadPolygon(line);

					for (int index = 0; index < polygon.PointCount; index++)
					{
						if (polygon.GetPoint(index) > this.numberOfVertices)
						{
							this.exception = "Error in input file on line " + this.actualLine + " - index is greater then number of vertices.";
							throw new Zcu.Mve.Core.MveException(null);
						}
					}

					this.dataSet.AddPolygon(i, polygon);

					if (++i == numberOfCells)              // Number of triangles reached - exit.
					{
						break;
					}
				}
			}

			if (i != this.numberOfCells)
			{
				this.exception = "Error in input file - number of cells and/or vertices doesn't match.";
				throw new Zcu.Mve.Core.MveException(null);
			}
		}
	}
}
