using System;
using System.IO;
using System.Text;
using System.Collections;

namespace Zcu.Gsvd.MeshLS
{
	/// <summary>
	/// Class for reading text STL file format.
	/// </summary>
	public class ReaderStl
	{
		private int numberOfCells = 0;
		private int numberOfVertices  = 0;

		private DataSet dataSet;

		private ArrayList cellNormalsArray = new ArrayList();
		private ArrayList cellArray        = new ArrayList();

		private Hashtable verticesHash     = new Hashtable();

		private int actualLine = 0;                        // Actual line of readed file.

		/// <summary>
		/// Reads input file.
		/// </summary>
		/// <param name="fileName">Full filename.</param>
		/// <returns>Filled object of DataSet class.</returns>
		public DataSet ReadFile(string fileName)
		{
			StreamReader streamReader = new StreamReader(fileName);

			string line = null;
			int    tmpCells = 0;

			try
			{
				while ((line = streamReader.ReadLine()) != null)
				{
					this.actualLine++;

					line = line.Trim();
					line = line.ToUpper();

					if (line.StartsWith("FACET NORMAL"))
					{                                      // Definition of cell normal.
						this.cellNormalsArray.Add(this.ReadNormal3D(line));
					}                                      // Adds normal for readed cell.

					else if (line.StartsWith("OUTER LOOP"))
					{                                      // Definition of cell vertices.
						tmpCells++;
						this.ReadTriangleVertex(ref streamReader);
					}                                      // Reads cell vertices.
				}
			}
			catch
			{
				streamReader.Close();
				throw new Zcu.Mve.Core.MveException("Error in input file on line " + this.actualLine + ".");
			}

			streamReader.Close();

			this.cellNormalsArray.TrimToSize();
			this.numberOfCells = this.cellNormalsArray.Capacity;
			                                               // Getting number of triangles (faces).

			if (this.numberOfCells != tmpCells)
			{                                              // Checks data consistency.
				streamReader.Close();
				throw new Zcu.Mve.Core.MveException("Error in input file - different number of triangles and triangles normals.");
			}

			this.dataSet = new DataSet(this.numberOfVertices, this.numberOfCells, 0);
                                                           // Creates new DataSet object.
			try
			{
				writeDataToDataSet();
			}
			catch
			{
				streamReader.Close();
				throw new Zcu.Mve.Core.MveException("Unexpected error in input file. Check input file structure.");
			}

			return this.dataSet;
		}

		/// <summary>
		/// Reads one vertex in input string.
		/// </summary>
		/// <param name="line">Input string.</param>
		/// <returns>Readed vertex (3D point).</returns>
		private Zcu.Mve.Visualization.Point3D ReadPoint3D(string line) 
		{
			string[] elements;
			int elementsLength = 0;

			System.Globalization.NumberFormatInfo nfi = new System.Globalization.NumberFormatInfo();

			nfi.CurrencyDecimalSeparator = ".";            // Setting of decimal separator.

			elements = line.Split(" :\t".ToCharArray());

			for (int i = 0; i < elements.Length; i++)      // Delete elements that contains spaces only.
			{
				if (elements[i].Trim() != "")
				{
					elements[elementsLength++] = elements[i];
				}
			}

			return new Zcu.Mve.Visualization.Point3D(Double.Parse(elements[elementsLength - 3], nfi),
				                                     Double.Parse(elements[elementsLength - 2], nfi),
				                                     Double.Parse(elements[elementsLength - 1], nfi));
		}

		/// <summary>
		/// Reads one normal in input string.
		/// </summary>
		/// <param name="line">Input string.</param>
		/// <returns>Readed normal (3D normal).</returns>
		private Zcu.Mve.Visualization.Normal3D ReadNormal3D(string line) 
		{
			string[] elements;
			int elementsLength = 0;

			System.Globalization.NumberFormatInfo nfi = new System.Globalization.NumberFormatInfo();

			nfi.CurrencyDecimalSeparator = ".";            // Setting of decimal separator.

			elements = line.Split(" :\t".ToCharArray());

			for (int i = 0; i < elements.Length; i++)      // Delete elements that contains spaces only.
			{
				if (elements[i].Trim() != "")
				{
					elements[elementsLength++] = elements[i];
				}
			}

			return new Zcu.Mve.Visualization.Normal3D(Double.Parse(elements[elementsLength - 3], nfi),
				                                      Double.Parse(elements[elementsLength - 2], nfi),
				                                      Double.Parse(elements[elementsLength - 1], nfi));
		}

		/// <summary>
		/// Reads tree vertices of triangle from input file.
		/// </summary>
		/// <param name="streamReader">Reference to input file.</param>
		private void ReadTriangleVertex(ref StreamReader streamReader)
		{
			Zcu.Mve.Visualization.Point3D  vertex1;
			Zcu.Mve.Visualization.Point3D  vertex2;
			Zcu.Mve.Visualization.Point3D  vertex3;
			Zcu.Mve.Visualization.Triangle triangle;

			string line;

			line = streamReader.ReadLine().Trim();         // Reads first vertex.
			this.actualLine++;
			vertex1 = ReadPoint3D(line);
			
			line = streamReader.ReadLine().Trim();         // Reads second vertex.
			this.actualLine++;
			vertex2 = ReadPoint3D(line);
			
			line = streamReader.ReadLine().Trim();         // Reads third vertex.
			this.actualLine++;
			vertex3 = ReadPoint3D(line);

			triangle = StoreVertexAndGetTriangle(vertex1, vertex2, vertex3);
			                                               // Call method for creating cell indexes.
			this.cellArray.Add(triangle);                  // Adds cell to array.
		}

		/// <summary>
		/// Finds indexes of triangle's vertices.
		/// </summary>
		/// <param name="vertex1">First vertex of triangle.</param>
		/// <param name="vertex2">Second vertex of triangle.</param>
		/// <param name="vertex3">Third vertex of triangle.</param>
		/// <returns>Indexes definition of triangle.</returns>
		private Zcu.Mve.Visualization.Triangle StoreVertexAndGetTriangle(Zcu.Mve.Visualization.Point3D vertex1,
			                                                             Zcu.Mve.Visualization.Point3D vertex2,
			                                                             Zcu.Mve.Visualization.Point3D vertex3)
		{
			Zcu.Mve.Visualization.Triangle triangle = new Zcu.Mve.Visualization.Triangle(0, 0, 0);

			if (this.numberOfVertices == 0)                 // Store first three vertices.
			{
				this.verticesHash.Add(vertex1, 0);
				this.verticesHash.Add(vertex2, 1);
				this.verticesHash.Add(vertex3, 2);
				this.numberOfVertices = 3;

				triangle.V1 = 0;                           // Assign idexes to first triangle.
				triangle.V2 = 1;
				triangle.V3 = 2;

				return triangle;
			}

			if (this.verticesHash.ContainsKey(vertex1))    // Checks if exist first vertex in vertices array.
			{                                              // True - assigns index to triangle.
				triangle.V1 = (int)this.verticesHash[vertex1];
			}
			else
			{                                              // False - adds new vertex to array.
				this.verticesHash.Add(vertex1, this.numberOfVertices);
				triangle.V1 = this.numberOfVertices;       // Assigns index (current number of vertices) to triangle.
				this.numberOfVertices++;
			}

			if (this.verticesHash.ContainsKey(vertex2))    // Same as first vertex.
			{
				triangle.V2 = (int)this.verticesHash[vertex2];
			}
			else
			{
				this.verticesHash.Add(vertex2, this.numberOfVertices);
				triangle.V2 = this.numberOfVertices;
				this.numberOfVertices++;
			}

			if (this.verticesHash.ContainsKey(vertex3))    // Same as first vertex.
			{
				triangle.V3 = (int)this.verticesHash[vertex3];
			}
			else
			{
				this.verticesHash.Add(vertex3, this.numberOfVertices);
				triangle.V3 = this.numberOfVertices;
				this.numberOfVertices++;
			}

			return triangle;
		}

		/// <summary>
		/// Adds local readed data to DataSet.
		/// </summary>
		private void writeDataToDataSet()
		{
			int   index = 0;

			foreach (Zcu.Mve.Visualization.Point3D point3d in this.verticesHash.Keys)
			{                                              // Adds vertices.
				this.dataSet.AddVertex((int)this.verticesHash[point3d] , point3d);
			}

			index = 0;
			foreach (Zcu.Mve.Visualization.Normal3D normal3d in this.cellNormalsArray)
			{                                              // Adds cell normals.
				this.dataSet.AddCellNormal(index, normal3d);
				index++;
			}

			index = 0;
			foreach (Zcu.Mve.Visualization.Triangle triangle in this.cellArray)
			{                                              // Adds cells.
				this.dataSet.AddTriangle(index, triangle);
				index++;
			}
		}
	}
}
